package Plugins::MySQL;
use DBI;
use Config::General;

my $conf = new Config::General(
    -ConfigFile      => $config::plugins_dir . '/MySQL.conf',
    -InterPolateVars => TRUE,
    -AutoLaunder     => TRUE
);
my %cfg = $conf->getall;

my @att_list = split(/,/, $cfg{'attribute_list'});
foreach $att (@att_list) {
    $store_att{$att}++;
}

$VERSION  = 1.0;
$this_pkg = (caller(0))[0];

######################################################################
# Extract the uid part from DN
######################################################################
sub get_uid {
    my ($dn) = @_;
    my ($uid, $tmp);

    ($uid, $tmp) = split(/,/, $dn);
    ($tmp, $uid) = split(/=/, $uid);
    return $uid;
}

##################################################################
#
##################################################################
sub my_connect {
    my $dbh = DBI->connect("dbi:mysql:database=" . $cfg{'dbname'} . ";host=" . $cfg{'dbhost'}, $cfg{'dbuser'}, $cfg{'dbpwd'}, { 'RaiseError' => 0 });

    return $dbh;
}

##################################################################
#  Makes a select query which sould return only one row
#  SELECT $fields FROM $tables WHERE $where_clause
##################################################################
sub my_select {
    my ($tables, $fields, $where_clause) = @_;
    my ($dbh, $query, $sth, $this_fun, $code, $msg, $out, @row);
    $this_fun = (caller(0))[3];
    $msg      = '';
    $code     = 0;

    AOSICFG::write_ws_log("|> $this_fun(in): tables=$tables;fields=$fields;where=$where_clause") if $config::aosi_debug_level > 2;
    if ($where_clause ne '') {
        $dbh = my_connect();
        if ($dbh) {
            $query = "SELECT $fields FROM $tables WHERE $where_clause";
            AOSICFG::write_ws_log("|> $this_fun: query=$query") if $config::aosi_debug_level > 3;
            $sth = $dbh->prepare($query) or do { $msg .= 'Error preparing statement: ' . $sth->errstr; $code += $sth->err; };
            if ($code == 0) {
                $sth->execute() or do { $msg .= 'Error executing statement: ' . $sth->errstr; $code += $sth->err; };
                if ($code == 0) {
                    $out = $sth->fetchall_arrayref or do { $msg .= 'Error fetching data: ' . $sth->errstr; $code += $sth->err; };
                    @row = @{ $out->[0] };
                }
            }
        }
        else {
            $code = -10201;
            $msg  = 'Could not connect to ' . $cfg{'dbname'} . ' on ' . $cfg{'dbhost'} . '!';
        }
    }
    else {
        @row  = ();
        $code = -10200;
        $msg  = 'No WHERE clause in SELECT statement!';
    }
    AOSICFG::write_ws_log("|> $this_fun(out): code=$code;msg=$msg;row=[" . join(", ", @row) . "]") if $config::aosi_debug_level > 2;
    return $code, $msg, @row;
}

##################################################################
#  Makes an insert
#  INSERT INTO $table($fields) VALUES($values)
##################################################################
sub my_insert {
    my ($table, $fields, $values) = @_;
    my ($dbh, $query, $sth, $this_fun, $code, $msg, $out);
    my (@fields, @values, $id);
    $this_fun = (caller(0))[3];
    $msg      = '';
    $code     = 0;

    AOSICFG::write_ws_log("|> $this_fun(in): table=$table;fields=$fields;values=$values") if $config::aosi_debug_level > 2;
    @fields = split(/,/, $fields);
    @values = split(/,/, $values);
    if ($#fields == $#values) {
        $dbh = my_connect();
        if ($dbh) {
            $query = "INSERT INTO $table($fields) VALUES($values)";
            AOSICFG::write_ws_log("|> $this_fun: query=$query") if $config::aosi_debug_level > 3;
            $sth = $dbh->prepare($query) or do { $msg .= 'Error preparing statement: ' . $sth->errstr; $code += $sth->err; };
            if ($code == 0) {
                $sth->execute() or do { $msg .= 'Error executing statement: ' . $sth->errstr; $code += $sth->err; };
                $id = $dbh->{'mysql_insertid'};
            }
        }
        else {
            $code = -10203;
            $msg  = 'Could not connect to ' . $cfg{'dbname'} . ' on ' . $cfg{'dbhost'} . '!';
        }
    }
    else {
        $out  = ();
        $code = -10202;
        $msg  = 'Fields # and values # for INSERT are not equal!';
    }
    AOSICFG::write_ws_log("|> $this_fun(out): code=$code;msg=$msg") if $config::aosi_debug_level > 2;
    return $code, $msg, $id;
}

##################################################################
#  Makes an update
#  UPDATE $table SET $field1 = $value1, $field2 = $value2 WHERE $where_clause
##################################################################
sub my_update {
    my ($table, $fields, $values, $where_clause) = @_;
    my ($dbh, $query, $sth, $this_fun, $code, $msg, $out);
    my (@fields, @values, $id, $field, @set);
    $this_fun = (caller(0))[3];
    $msg      = '';
    $code     = 0;

    AOSICFG::write_ws_log("|> $this_fun(in): table=$table;fields=$fields;values=$values") if $config::aosi_debug_level > 2;
    @fields = split(/,/, $fields);
    @values = split(/,/, $values);
    if ($#fields == $#values) {
        if ($where_clause ne '') {
            $dbh = my_connect();
            foreach $i (0 .. $#fields) {
                push(@set, $fields[$i] . ' = ' . $values[$i]);
            }
            if ($dbh) {
                $query = "UPDATE $table SET " . join(", ", @set) . " WHERE $where_clause";
                AOSICFG::write_ws_log("|> $this_fun: query=$query") if $config::aosi_debug_level > 3;
                $sth = $dbh->prepare($query) or do { $msg .= 'Error preparing statement: ' . $sth->errstr; $code += $sth->err; };
                if ($code == 0) {
                    $sth->execute() or do { $msg .= 'Error executing statement: ' . $sth->errstr; $code += $sth->err; };
                }
            }
            else {
                $code = -10206;
                $msg  = 'Could not connect to ' . $cfg{'dbname'} . ' on ' . $cfg{'dbhost'} . '!';
            }
        }
        else {
            $out  = ();
            $code = -10205;
            $msg  = 'No WHERE clause in UPDATE statement!';
        }
    }
    else {
        $out  = ();
        $code = -10204;
        $msg  = 'Fields # and values # for INSERT are not equal!';
    }
    AOSICFG::write_ws_log("|> $this_fun (out): code=$code;msg=$msg") if $config::aosi_debug_level > 2;
    return $code, $msg;
}

##################################################################
#   beforeAddUser
##################################################################
sub beforeAddUser {
    my ($status_hash, $ldap, $base_dn, $entry_in) = @_;
    my ($this_fun, $msg, $code, $out_panic, $where_clause);
    my ($new_dn, $temp, $new_uid, $uid);
    $this_fun  = (caller(0))[3];
    $msg       = '';
    $code      = 0;
    $out_panic = 0;

    AOSICFG::write_ws_log("|> $this_fun(in): ldap=$ldap;\nbase_dn=$base_dn;\nentry_in=\n" . Plugins::dump_entry($entry_in)) if $config::aosi_debug_level > 1;
    $new_dn = $entry_in->{'dn'};
    ($new_uid, $temp, $temp) = split(/,/, $new_dn);
    ($temp, $new_uid) = split(/=/, $new_uid);
    $where_clause = "UID = '" . $new_uid . "'";
    if ($cfg{'active_col'}) {
        $where_clause .= " AND " . $cfg{'active_col'} . " = 1";
    }
    ($code, $msg, $uid) = my_select($cfg{'users_table'}, "UID", $where_clause);
    AOSICFG::write_ws_log("|> $this_fun: uid=$uid") if $config::aosi_debug_level > 1;
    if ($uid ne '') {
        $code = -10210;
        $msg  = "User $new_uid allready exists!";

        #$out_panic = 1;
    }
    AOSICFG::write_ws_log("|> $this_fun(out): code=$code;msg=$msg;panic=$out_panic;") if $config::aosi_debug_level > 1;
    return $code, $msg, $out_panic;
}

##################################################################
#  Add an account
##################################################################
sub afterAddUser {
    my ($status_hash, $entry_in, $out_xml, $user) = @_;
    my ($this_fun, $code, $msg,       $out_panic, $before_code, $before_msg, $before_panic);
    my ($uid,      $dn,   $ent_attrs, $attr,      @fields,      @values,     $val);

    $this_fun  = (caller(0))[3];
    $msg       = '';
    $code      = 0;
    $out_panic = 0;

    AOSICFG::write_ws_log("|> $this_fun(in): user=$user;entry_in=$entry_in;out_xml=$out_xml;") if $config::aosi_debug_level > 1;

    ($code, $msg) = Plugins::parse_ldap_xml($out_xml);
    AOSICFG::write_ws_log("|> $this_fun: ldap_code=$code;ldap_msg=$msg;") if $config::aosi_debug_level > 2;

    $before_code  = $status_hash->{'MySQL'}->{'before'}->{'code'};
    $before_msg   = $status_hash->{'MySQL'}->{'before'}->{'msg'};
    $before_panic = $status_hash->{'MySQL'}->{'before'}->{'panic'};

    if ($before_code == 0 and $before_panic == 0) {
        if ($code == 0) {    # This one is from parse_ldap_xml
            $dn        = $entry_in->{'dn'};
            $uid       = get_uid($dn);
            $ent_attrs = $entry_in->{'attrs'};
            ($code, $msg) = my_insert($cfg{'users_table'}, "UID," . $cfg{'active_col'} . "," . $cfg{'changedby_col'}, "'$uid',1,'$user'");
            foreach $attr (@$ent_attrs) {    # Fetch some attributes
                if ($store_att{ $attr->{'att'} }) {
                    if (ref($attr->{'val'}) eq 'ARRAY') {
                        foreach $val (@{ $attr->{'val'} }) {
                            ($code, $msg) = my_insert($cfg{'values_table'}, "UID,NAME,VALUE," . $cfg{'active_col'}, "'$uid','" . $attr->{'att'} . "','$val',1");
                        }
                    }
                    else {
                        ($code, $msg) =
                          my_insert($cfg{'values_table'}, "UID,NAME,VALUE," . $cfg{'active_col'}, "'$uid','" . $attr->{'att'} . "','" . $attr->{'val'} . "',1");
                    }
                }
            }
        }
    }
    else {
        $code = -10220;
        $msg  = "Before action returned: panic=" . $before_panic . "; code=" . $before_code . "; message=" . $before_msg;
    }
    AOSICFG::write_ws_log("|> $this_fun (out): code=$code;msg=$msg") if $config::aosi_debug_level > 1;
    return $code, $msg, $out_panic;
}

##################################################################
# Check if an account exists
##################################################################
sub beforeDeleteUser {
    my ($status_hash, $ldap, $base_dn, $entry_in, $user) = @_;
    my ($this_fun, $msg,     $code, $out_panic);
    my ($del_dn,   $del_uid, $uid,  $where_clause);
    $this_fun  = (caller(0))[3];
    $msg       = '';
    $code      = 0;
    $out_panic = 0;

    AOSICFG::write_ws_log("|> $this_fun(in): user=$user;entry_in=$entry_in;") if $config::aosi_debug_level > 1;
    $del_dn       = $entry_in->{'dn'};
    $del_uid      = get_uid($del_dn);
    $where_clause = "UID = '" . $del_uid . "'";
    if ($cfg{'active_col'}) {
        $where_clause .= " AND " . $cfg{'active_col'} . " = 1";
    }
    ($code, $msg, $uid) = my_select($cfg{'users_table'}, "UID", $where_clause);
    AOSICFG::write_ws_log("|> $this_fun: uid=" . join(", ", @$uid)) if $config::aosi_debug_level > 1;
    if ($uid eq '') {
        $code = -10230;
        $msg  = "User $del_uid does not exists!";

        #$out_panic = 1;
    }
    AOSICFG::write_ws_log("|> $this_fun(out): code=$code;msg=$msg") if $config::aosi_debug_level > 1;
    return $code, $msg, $out_panic;
}

##################################################################
# Delete an account
##################################################################
sub afterDeleteUser {
    my ($status_hash, $entry_in, $out_xml, $user, $pwd) = @_;
    my ($this_fun, $msg, $code, $out_panic, $cur_code, $cur_msg);
    my ($del_dn, $temp, $del_uid, $uid, $where_clause);

    $this_fun = (caller(0))[3];
    $msg      = '';
    $code     = 0;

    AOSICFG::write_ws_log("|> $this_fun(in): user=$user;entry_in=$entry_in;out_xml=$out_xml;") if $config::aosi_debug_level > 1;

    ($code, $msg) = Plugins::parse_ldap_xml($out_xml);
    AOSICFG::write_ws_log("|> $this_fun: ldap_code=$code;ldap_msg=$msg;") if $config::aosi_debug_level > 2;

    $before_code  = $status_hash->{'MySQL'}->{'before'}->{'code'};
    $before_msg   = $status_hash->{'MySQL'}->{'before'}->{'msg'};
    $before_panic = $status_hash->{'MySQL'}->{'before'}->{'panic'};

    if ($before_code == 0 and $before_panic == 0) {
        if ($code == 0) {    # This one is from parse_ldap_xml
            $del_dn       = $entry_in->{'dn'};
            $del_uid      = get_uid($del_dn);
            $where_clause = "UID = '" . $del_uid . "'";
            if ($cfg{'active_col'}) {
                $where_clause .= " AND " . $cfg{'active_col'} . " = 1";
            }
            ($cur_code, $cur_msg) = my_update($cfg{'users_table'},  $cfg{'active_col'}, "0", $where_clause);
            ($code,     $msg)     = my_update($cfg{'values_table'}, $cfg{'active_col'}, "0", $where_clause);
            $code += $cur_code;
            $msg .= $cur_msg;
        }
    }
    else {
        $code = -10040;
        $msg  = "Before action returned: panic=" . $before_panic . "; code=" . $before_code . "; message=" . $before_msg;
    }
    AOSICFG::write_ws_log("|> $this_fun(out): code=$code;msg=$msg") if $config::aosi_debug_level > 1;
    return $code, $msg, $out_panic;
}

##################################################################
#  Check if adding an existig value or deleting a nonexisting value
##################################################################
# Same function for all actions
# 1 = beforeAddAttribute
# 2 = beforeModifyAttribute
# 3 = beforeDeleteAttribute
sub beforeChangeAttribute {
    my ($status_hash, $type, $ldap, $base_dn, $entry_in, $user) = @_;
    my ($this_fun, $msg, $code, $out_panic, $attrs);
    my ($uid, $dn, $search, $att, $val, $cur_code, $cur_msg, @entries, $user_uid, $realm);
    $this_fun  = (caller(0))[3];
    $msg       = '';
    $code      = 0;
    $out_panic = 0;

    AOSICFG::write_ws_log("|> $this_fun(in): tpye=$type;ldap=$ldap;base=$base_dn;user=$user;entry_in=$entry_in;") if $config::aosi_debug_level > 1;

    $dn = $entry_in->{'dn'};
    foreach $attrs (@{ $entry_in->{'attrs'} }) {
        $att = $attrs->{'att'};
        $val = $attrs->{'val'}->[0];
        AOSICFG::write_ws_log("|> $this_fun: att=$att;val=$val") if $config::aosi_debug_level > 2;
        $uid          = get_uid($dn);
        $where_clause = "UID = '" . $uid . "' AND " . $cfg{'active_col'} . " = 1 AND NAME = '" . $att . "' AND VALUE = '" . $val . "'";
        ($cur_code, $cur_msg, $uid) = my_select($cfg{'values_table'}, "UID", $where_clause);

        # Adding an existing value
        if (($type == 1) and ($uid ne '')) {
            $cur_code = -10250;
            $cur_msg  = "Attribute $att with value $val already exists!";

            #$out_panic = 1;
        }

        # Deleting a nonexisting value
        if (($type == 3) and ($uid eq '')) {
            $cur_code = -10251;
            $cur_msg  = "There is no attribute $att with value $val!";

            #$out_panic = 1;
        }

        $code += abs($cur_code);
        $msg .= '<plugin_result name="'.$att.'" code="' . $cur_code . '">' . $cur_msg . '</plugin_result>';
    }
    $code = -$code;
    AOSICFG::write_ws_log("|> $this_fun(out): code=$code;msg=$msg") if $config::aosi_debug_level > 1;
    return $code, $msg, $out_panic;
}

##################################################################
#  Change attributes (userPassword in AD is in fact unicodePwd)
##################################################################
# Same function for all actions
# 1 = afterAddAttribute
# 2 = afterModifyAttribute
# 3 = afterDeleteAttribute
sub afterChangeAttribute {
    my ($status_hash, $type, $entry_in, $out_xml, $user, $pwd) = @_;
    my ($this_fun, $out_panic, $msg, $attrs, $tmp_code, $tmp_msg, $tmp_uid);
    my ($uid, $dn, $att, $val, $old_val, @old_vals, @new_vals, $cur_code, $cur_msg, $code, $where_clause);
    $this_fun  = (caller(0))[3];
    $msg       = '';
    $code      = 0;
    $out_panic = 0;

    AOSICFG::write_ws_log("|> $this_fun(in): tpye=$type;ldap=$ldap;base=$base_dn;user=$user;entry_in=$entry_in;") if $config::aosi_debug_level > 1;
    ($code, $msg) = Plugins::parse_ldap_xml($out_xml);
    AOSICFG::write_ws_log("|> $this_fun: ldap_code=$code;ldap_msg=$msg;") if $config::aosi_debug_level > 2;

    $before_code  = $status_hash->{'MySQL'}->{'before'}->{'code'};
    $before_msg   = $status_hash->{'MySQL'}->{'before'}->{'msg'};
    $before_panic = $status_hash->{'MySQL'}->{'before'}->{'panic'};
    if ($before_code == 0 and $before_panic == 0) {
        $msg = '';
        $dn  = $entry_in->{'dn'};
        foreach $attrs (@{ $entry_in->{'attrs'} }) {
            $att = $attrs->{'att'};
            $val = $attrs->{'val'}->[0];
            if ($store_att{$att} > 0) {
                AOSICFG::write_ws_log("|> $this_fun: att=$att;val=$val") if $config::aosi_debug_level > 2;
                $uid = get_uid($dn);
                $where_clause = "UID = '" . $uid . "' AND " . $cfg{'active_col'} . " = 1 AND NAME = '" . $att . "' AND VALUE = '" . $val . "'";
                ($tmp_code, $tmp_msg, $tmp_uid) = my_select($cfg{'values_table'}, "UID", $where_clause);
                if($tmp_uid ne ''){
                    if ($type == 2) {    # Modify = Delete_all + Add
                        $where_clause = "UID = '" . $uid . "' AND " . $cfg{'active_col'} . " = 1";
                        ($cur_code, $cur_msg) = my_update($cfg{'values_table'}, $cfg{'active_col'}, "0", $where_clause);
                        ($tmp_code, $tmp_msg) = my_insert($cfg{'values_table'}, "UID,NAME,VALUE," . $cfg{'active_col'}, "'$uid','" . $att . "','" . $val . "',1");
                        $cur_code += $tmp_code;
                        $cur_msg .= $tmp_msg;
                    }
                    elsif ($type == 3) {    # Delete
                        if ($val ne '') {
                            $where_clause = "UID = '" . $uid . "' AND " . $cfg{'active_col'} . " = 1 AND NAME = '" . $att . "' AND VALUE = '" . $val . "'";
                            ($cur_code, $cur_msg, $uid) = my_select($cfg{'values_table'}, "UID", $where_clause);
                            if ($uid ne '') {
                                $where_clause = "UID = '" . $uid . "' AND " . $cfg{'active_col'} . " = 1 AND NAME = '" . $att . "' AND VALUE = '" . $val . "'";
                                ($cur_code, $cur_msg) = my_update($cfg{'values_table'}, $cfg{'active_col'}, "0", $where_clause);
                            }
                        }
                        else {
                            $where_clause = "UID = '" . $uid . "' AND " . $cfg{'active_col'} . " = 1";
                            ($cur_code, $cur_msg) = my_update($cfg{'values_table'}, $cfg{'active_col'}, "0", $where_clause);
                        }
                    }
                }
                if ($type == 1) {    # Add
                    ($cur_code, $cur_msg) = my_insert($cfg{'values_table'}, "UID,NAME,VALUE," . $cfg{'active_col'}, "'$uid','" . $att . "','" . $val . "',1");
                }
                $code += abs($cur_code);
                $msg .= '<plugin_result name="'.$att.'" code="' . $cur_code . '">' . $cur_msg . '</plugin_result>';
            }
        }
        $code = -$code;
    }
    else {
        $code = -10260;
        $msg  = "Before action returned: panic=" . $before_panic . "; code=" . $before_code . "; message=" . $before_msg;
    }

    AOSICFG::write_ws_log('|> ' . $this_fun . '(out): code=' . $code . ';msg=' . $msg) if $config::aosi_debug_level > 1;
    return $code, $msg, $out_panic;
}

1;
